/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "server.h"
#include "assert/assert.h"
#include "console/console.h"

server_t SERVER;

void serverInit(const serverinit_t init) {
  memset(&SERVER, 0, sizeof(server_t));
  SERVER.state = SERVER_STATE_STARTING;
  SERVER.type = init.type;

  // Create the server
  switch(init.type) {
    case SERVER_TYPE_ONLINE:
      SERVER.online.socket = socket(AF_INET, SOCK_STREAM, 0);
      SERVER.online.address.sin_family = AF_INET;
      SERVER.online.address.sin_addr.s_addr = INADDR_ANY;
      SERVER.online.address.sin_port = htons(init.online.port);

      // Bind
      if(bind(
        SERVER.online.socket,
        (struct sockaddr *)&SERVER.online.address,
        sizeof(SERVER.online.address)
      ) < 0) {
        assertUnreachable("Failed to bind server.");
      }

      // Begin listening
      if(listen(SERVER.online.socket, SERVER_CLIENT_COUNT_MAX) < 0) {
        assertUnreachable("Failed to listen on server.");
      }

      // Set the socket to non-blocking mode
      int flags = fcntl(SERVER.online.socket, F_GETFL, 0);
      fcntl(SERVER.online.socket, F_SETFL, flags | O_NONBLOCK);

      break;
    
    default:
      assertUnreachable("Invalid server type.");
      break;
  }

  // Start thread
  pthread_create(&SERVER.thread, NULL, serverThread, NULL);

  // Wait for server to start
  while(SERVER.state == SERVER_STATE_STARTING) usleep(1000);
}

void * serverThread(void *arg) {
  assertNull(arg, "Server thread does not accept arguments.");
  
  // Server now running
  SERVER.state = SERVER_STATE_RUNNING;

  fd_set readfds;
  int32_t activity;
  int32_t error;
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 1000;

  // Start server
  while(SERVER.state == SERVER_STATE_RUNNING) {
    FD_ZERO(&readfds);
    FD_SET(SERVER.online.socket, &readfds);

    activity = select(
      SERVER.online.socket + 1,
      &readfds,
      NULL,
      NULL,
      &timeout
    );

    if(activity < 0) assertUnreachable("Select error");

    if(!FD_ISSET(SERVER.online.socket, &readfds)) {
      continue;
    }

    // Accept incoming connections
    int32_t clientSockDesc = accept(SERVER.online.socket, NULL, NULL);
    if(clientSockDesc < 0) {
      printf("Failed to accept connection\n");
      continue;
    }

    // Now, is the server full?
    if(SERVER.clientCount >= SERVER_CLIENT_COUNT_MAX) {
      consolePrint("Server is full, rejecting connection.");
      // TODO: Write a reason to client.
      close(clientSockDesc);
      continue;
    }

    // Create server client for handling
    serverclient_t *client = &SERVER.clients[SERVER.clientCount];
    serverClientInit(client, (serverclientinit_t){
      .socket = clientSockDesc
    });
    SERVER.clientCount++;
  }

  SERVER.state = SERVER_STATE_STOPPED;
  return NULL;
}

void serverDispose() {
  // Disconnect clients
  for(uint8_t i = 0; i < SERVER.clientCount; i++) {
    serverClientDispose(&SERVER.clients[i]);
  }
  SERVER.clientCount = 0;

  // Stop server thread.
  if(SERVER.state == SERVER_STATE_RUNNING) {
    // Request thread to stop
    SERVER.state = SERVER_STATE_STOPPING;

    // Wait for server to stop
    while(SERVER.state == SERVER_STATE_STOPPING) usleep(1000);
    pthread_join(SERVER.thread, NULL);

    // Close the socket
    switch(SERVER.type) {
      case SERVER_TYPE_ONLINE:
        close(SERVER.online.socket);
        break;
      
      default:
        assertUnreachable("Invalid server type.");
        break;
    }
  }
}