/**
 * Copyright (c) 2025 Dominic Masters
 * 
 * This software is released under the MIT License.
 * https://opensource.org/licenses/MIT
 */

#include "serverclient.h"
#include "assert/assert.h"
#include "console/console.h"

void serverClientInit(serverclient_t *client, const serverclientinit_t init) {
  assertNotNull(client, "Client cannot be NULL.");
  client->socket = init.socket;
  consolePrint("Accepted client %i", client->socket);

  client->state = SERVER_CLIENT_STATE_CONNECTING;

  // Set socket to non-blocking
  int flags = fcntl(client->socket, F_GETFL, 0);
  fcntl(client->socket, F_SETFL, flags | O_NONBLOCK);

  // Start thread
  pthread_create(&client->thread, NULL, serverClientThread, client);

  // Wait for client to connect
  while(client->state == SERVER_CLIENT_STATE_CONNECTING) usleep(1000);
}

void * serverClientThread(void *arg) {
  serverclient_t *client = (serverclient_t*)arg;
  assertNotNull(client, "Client cannot be NULL.");

  // Set the client state to connected
  client->state = SERVER_CLIENT_STATE_CONNECTED;

  // Send hello to client, as well as the version number.
  {
    uint8_t hello[16];
    memset(hello, 0, sizeof(hello));
    memcpy(hello, "Dusk", 4);
    strcpy((char*)&hello[4], DUSK_VERSION);
    
    _serverClientSend(client, hello, sizeof(hello));
  }

  // Client should respond "OK"
  {
    uint8_t response[2];
    _serverClientReceive(client, response, sizeof(response));
    assertTrue(
      response[0] == 'O' &&
      response[1] == 'K',
      "Client did not respond with OK."
    );
  }
  
  while(client->state == SERVER_CLIENT_STATE_CONNECTED) {
    // Do nothing for now.
    usleep(1000);
  }

  // Disconnect
  client->state = SERVER_CLIENT_STATE_DISCONNECTED;
  return NULL;
}

void _serverClientSend(
  serverclient_t *client,
  const uint8_t *data,
  const size_t size
) {
  assertNotNull(client, "Client cannot be NULL.");
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");
}

void _serverClientReceive(
  serverclient_t *client,
  uint8_t *data,
  const size_t size
) {
  assertNotNull(client, "Client cannot be NULL.");
  assertNotNull(data, "Data cannot be NULL.");
  assertTrue(size > 0, "Size cannot be zero.");
  
  recv(client->socket, data, size, 0);
}

void serverClientDispose(serverclient_t *client) {
  assertNotNull(client, "Client cannot be NULL.");

  if(client->state == SERVER_CLIENT_STATE_CONNECTED) {
    // Let the thread know we are disconnecting
    client->state = SERVER_CLIENT_STATE_DISCONNECTING;

    // Wait for the thread to finish
    while(client->state == SERVER_CLIENT_STATE_DISCONNECTING) usleep(1000);
    pthread_join(client->thread, NULL);
  }
}